/* -*- c++ -*- */

/*

Part of DiyouPCBFirm - PCB blue-ray laser printer firmware based on Marlin, GRBL and HPGLxx

Copyright (C) 2013 V.Ayllon
Copyright (C) 2011 Camiel Gubbels / Erik van der Zalm
Copyright (c) 2011-2012 Sungeun K. Jeon
Copyright (c) 2011 Jens Geisler
Copyright (c) 2009-2011 Simen Svale Skogsrud

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "DiyouPCBFirm.h"
#include "planner.h"
#include "stepper.h"
#include "motion_control.h"
#include "language.h"
#include "hpgl_parser.h"
#include "laser_control.h"
#include "focus.h"
#include "backlash.h" 

#define VERSION_STRING  "1.0"

//===========================================================================
//=============================private variables=============================
//===========================================================================
static char cmdbuffer[BUFSIZE][MAX_CMD_SIZE];
static int bufindr = 0;
static int bufindw = 0;
static int buflen = 0;
static int serial_count = 0;

system_t sys; 

//===========================================================================
//=============================ROUTINES=============================
//===========================================================================

void serial_echopair_P(const char *s_P, float v)
{ serialprintPGM(s_P); SERIAL_ECHO(v); }
void serial_echopair_P(const char *s_P, double v)
{ serialprintPGM(s_P); SERIAL_ECHO(v); }
void serial_echopair_P(const char *s_P, unsigned long v)
{ serialprintPGM(s_P); SERIAL_ECHO(v); }


extern "C"{
	extern unsigned int __bss_end;
	extern unsigned int __heap_start;
	extern void *__brkval;

	int freeMemory() {
		int free_memory;

		if((int)__brkval == 0)
			free_memory = ((int)&free_memory) - ((int)&__bss_end);
		else
			free_memory = ((int)&free_memory) - ((int)__brkval);

		return free_memory;
	}
}


/**
** init_printer() - Initialization
**/
void init_printer()
{
	config_reset_default();
    hpgl_init();     // Init parser
    laser_init();    // Init laser system
	turn_laser(OFF); // Turn of the laser
	planInit();		// Initialize planner;
	st_init();    // Initialize stepper, this enables interrupts!
	sei();


	SERIAL_PROTOCOLLNPGM("start");
	SERIAL_ECHO_START;

	SERIAL_ECHOPGM(MSG_NAME);
	SERIAL_ECHOLNPGM(VERSION_STRING);
#ifdef STRING_VERSION_CONFIG_H
#ifdef STRING_CONFIG_H_AUTHOR
	SERIAL_ECHO_START;
	SERIAL_ECHOPGM(MSG_CONFIGURATION_VER);
	SERIAL_ECHOPGM(STRING_VERSION_CONFIG_H);
	SERIAL_ECHOPGM(MSG_AUTHOR);
	SERIAL_ECHOLNPGM(STRING_CONFIG_H_AUTHOR);
	SERIAL_ECHOPGM("Compiled: ");
	SERIAL_ECHOLNPGM(__DATE__);
#endif
#endif
	SERIAL_ECHO_START;
	SERIAL_ECHOPGM(MSG_FREE_MEMORY);
	SERIAL_ECHO(freeMemory());
	SERIAL_ECHOPGM(MSG_PLANNER_BUFFER_BYTES);
	SERIAL_ECHOLN((int)sizeof(block_t)*BLOCK_BUFFER_SIZE);
	// Go home
	homeAllAxis(false);
	// Send ok to start transmisión.
	disable_motors();
	clearToSend();
}
// Executes run-time commands, when required. This is called from various check points in the main
// program, primarily where there may be a while loop waiting for a buffer to clear space or any
// point where the execution time from the last check point may be more than a fraction of a second.
// This is a way to execute runtime commands asynchronously.
void execute_runtime_cmd()
{
  if (sys.execute) { // Enter only if any bit flag is true
    uint8_t rt_exec = sys.execute; // Avoid calling volatile multiple times
     
    // Execute system reset. 
    if (rt_exec & EXEC_RESET) {
      init_printer();
	  bit_false(sys.execute,EXEC_RESET);
	  sys.state = STATE_ONLINE;  
      return; // Nothing else to do but exit.
    }
    
    // Execute and serial print status
    if (rt_exec & EXEC_STATUS_REPORT) { 
      report_status();
      bit_false(sys.execute,EXEC_STATUS_REPORT);
    }
    
	// Initiate stepper feed hold
	if (rt_exec & EXEC_FEED_HOLD) {
		if (sys.state == STATE_ONLINE) {
			sys.state = STATE_HOLD;
		}
		bit_false(sys.execute,EXEC_FEED_HOLD);
	}
    // If restart check if it´s holding or held
	// if holding wait till held and wake up
	// the stepper system.
	if (rt_exec & EXEC_CYCLE_START) { 
		if (sys.state == STATE_HELD) {
			sys.state = STATE_ONLINE;
			st_wake_up();
		}
		bit_false(sys.execute,EXEC_CYCLE_START);
	}
  }
}  
// Method to ready the system to reset by setting the runtime reset command and killing any
// active processes in the system. This also checks if a system reset is issued while Grbl
// is in a motion state. If so, kills the steppers and sets the system alarm to flag position
// lost, since there was an abrupt uncontrolled deceleration. Called at an interrupt level by
// runtime abort command and hard limits. So, keep to a minimum.
void reset_printer(){
	// Turn laser off
	// Stop steppers
	turn_laser(OFF);
	quickStop();
	sys.execute |= EXEC_RESET;
}

/**
** report_status() - Reports the status of the printer
**/
void report_status() {

	// Report current machine state & position
	switch (sys.state) {
	case STATE_IDLE: SERIAL_ECHOPGM(MSG_STATE_IDLE); break;
	case STATE_ONLINE: SERIAL_ECHOPGM(MSG_STATE_ONLINE); break;
	case STATE_HOLD: SERIAL_ECHOPGM(MSG_STATE_HOLD); break;
	case STATE_HELD: SERIAL_ECHOPGM(MSG_STATE_HELD); break;
	default: SERIAL_ECHOPGM(MSG_STATE_UNKNOW); break;
	}

	SERIAL_ECHOPGM("- Pos: ");
	printFloat(hpgl_user_to_plotter(sys.current_position[X_AXIS],X_AXIS),0);
	SERIAL_ECHOPGM(",");
	printFloat(hpgl_user_to_plotter(sys.current_position[Y_AXIS],Y_AXIS),0);
	SERIAL_ECHOPGM(">\r\n");
	// Report feed rate
	SERIAL_ECHOPGM(MSG_FEED_RATE);
	printFloat(sys.current_feed_rate,0);
	SERIAL_ECHOPGM("(mm/sec) -");
	SERIAL_ECHOPGM(MSG_FREE_MEMORY);
	SERIAL_ECHO(freeMemory());
	SERIAL_ECHOPGM("(bytes)\r\n");
	SERIAL_ECHOPGM(MSG_BACKLASH);
	printFloat(sys.backlash[X_AXIS],2);
	SERIAL_ECHOPGM(",");
	printFloat(sys.backlash[Y_AXIS],2);
	SERIAL_ECHOPGM("\r\n");
}

// Process and report status one line of incoming serial data. Performs an initial filtering
// by removing spaces and comments and capitalizing all letters.
void getCommand()
{
  uint8_t c;

  while((c = serial_read()) != SERIAL_NO_DATA) {
	  if ((c == ';') || (c == '\n') || (c == '\r')) { // End of command or line reached

		  if (serial_count > 0) {// Line is complete. Then execute!
			  cmdbuffer[bufindw][serial_count] = '\0'; // Terminate string
			  bufindw = (bufindw + 1)%BUFSIZE;
			  buflen += 1;
			  serial_count = 0;
			  return;
		  }
		  // Empty line
		  serial_count = 0;
		  return;
	  } else {
		  if (c <= ' ') { 
			  // Throw away whitepace and control characters
		  } else if (serial_count >= MAX_CMD_SIZE-1) {
			  // Report line buffer overflow and reset
			  SERIAL_ERRORLNPGM(MSG_OVERFLOW);
			  serial_count = 0;
			  return;
		  } else if (c >= 'a' && c <= 'z') { // Upcase lowercase
			  cmdbuffer[bufindw][serial_count++] = c-'a'+'A';
		  } else {
			  cmdbuffer[bufindw][serial_count++] = c;
		  }
	  }
  }
}


/**
**process_hpgl_commands() - Process a HPGL command from the buffer
**/
void process_hpgl_commands(char *line) 
{ 
	uint16_t command;

	command = hpgl_parser_run(line);

	switch(command) {
	case STCMD:
		sys.execute |= EXEC_STATUS_REPORT;
		break;
	case HYCMD:
		if (numpar[0] >= 0){
			sys.backlash[X_AXIS]= numpar[0];
		}
		if (numpar[1] >= 0){
			sys.backlash[Y_AXIS]= numpar[1];
		}
		break;
	case LCMD:
		if (numpar[0] == 1){
			set_laser_power(sys.current_laser_power);
			turn_laser(ON);
		}
		else {
			turn_laser(OFF);
		}
		break;
	case PCMD:
		if (numpar[0] > 0 && numpar[0] < MAX_LASER_POWER){
			set_laser_power((uint8_t)numpar[0]);
			sys.current_laser_power = (uint8_t)numpar[0];
		}
		break;
	case HP_PU:
		sys.current_laser_state = OFF;
		break;
	case HP_PD:
		sys.current_laser_state = ON;
		break;
	case HP_PA:
		sys.target[X_AXIS] = (float)numpar[X_COORD];
		sys.target[Y_AXIS] = (float)numpar[Y_COORD];
		prepare_line_move();
		break;
	case HP_AA:
		prepare_arc_move();
		break;
	case HP_IN:		// Home & init
		homeAllAxis(true);
		break;
	case HP_VS:
		if (numpar[0] > 0 && numpar[0] <= MAX_FEEDRATE){
			sys.current_feed_rate = numpar[0];
		}
		break;
	case HP_SP:		// Select pen 
	case HP_IP:		// Init point
	case HP_SC:		// Scale all of them processed inside HPGL parser
		break;
	default: 
		break;
	}

	clearToSend();
}


#define DEFINE_PGM_READ_ANY(type, reader)		\
	static inline type pgm_read_any(const type *p)	\
{ return pgm_read_##reader##_near(p); }

DEFINE_PGM_READ_ANY(float,       float);
DEFINE_PGM_READ_ANY(signed char, byte);

#define XYZ_CONSTS_FROM_CONFIG(type, array, CONFIG)	\
	static const PROGMEM type array##_P[NUM_AXIS] =		\
{ X_##CONFIG, Y_##CONFIG};		\
	static inline type array(int axis)			\
{ return pgm_read_any(&array##_P[axis]); }

XYZ_CONSTS_FROM_CONFIG(float, base_home_pos,   HOME_POS);
XYZ_CONSTS_FROM_CONFIG(float, max_length,      MAX_LENGTH);
XYZ_CONSTS_FROM_CONFIG(float, home_retract_mm, HOME_RETRACT_MM);
XYZ_CONSTS_FROM_CONFIG(signed char, home_dir,  HOME_DIR);

/**
** axisIsAtHome() - home one axis
**/
static void axisIsAtHome(int axis) {
	sys.current_position[axis] = base_home_pos(axis);
}
/**
** homeAxis() - home one axis
**/
static void homeaxis(int axis) {
#define HOMEAXIS_DO(LETTER) \
	((LETTER##_MIN_PIN > -1 && LETTER##_HOME_DIR==-1))

	if (axis==X_AXIS ? HOMEAXIS_DO(X) :
		axis==Y_AXIS ? HOMEAXIS_DO(Y) :0) {
			sys.current_position[axis] = 0;
			sys.backlash[axis]= 0.0f;
			sys.current_dir[axis] = -1;

			planSetPosition(sys.current_position[X_AXIS], sys.current_position[Y_AXIS]);
			sys.target[axis] = 1.5 * max_length(axis) * home_dir(axis);
			planBufferLine(sys.target[X_AXIS], sys.target[Y_AXIS], HOMING_FEEDRATE,OFF, 0,0,0);
			st_synchronize();

			sys.current_position[axis] = 0;
			planSetPosition(sys.current_position[X_AXIS], sys.current_position[Y_AXIS]);
			sys.target[axis] = -home_retract_mm(axis) * home_dir(axis);
			planBufferLine(sys.target[X_AXIS], sys.target[Y_AXIS], HOMING_FEEDRATE,OFF, 0,0,0);
			st_synchronize();

			sys.target[axis] = 2*home_retract_mm(axis) * home_dir(axis);
			planBufferLine(sys.target[X_AXIS], sys.target[Y_AXIS], HOMING_FEEDRATE/6,OFF, 0,0,0);
			st_synchronize();

			axisIsAtHome(axis);					
			sys.target[axis] = sys.current_position[axis];
			endstops_hit_on_purpose();
			sys.current_dir[axis] = 1;
	}
}
#define HOMEAXIS(LETTER) homeaxis(LETTER##_AXIS)
/**
** homeAllAxis() - Home all axis
**/
void homeAllAxis(bool focus) {

	turn_laser(OFF);

	enable_endstops(true);

	HOMEAXIS(Y);

	HOMEAXIS(X);

    planSetPosition(sys.current_position[X_AXIS], sys.current_position[Y_AXIS]);
	enable_endstops(false);

	manageInactivity();

	// Now we are at home position

	if (focus){
		SERIAL_ECHOLNPGM(MSG_FOCUSING);
		//
		// Move to focus offset position
		//
		sys.target[X_AXIS] = sys.current_position[X_AXIS] + FOCUS_OFFSET_X;
		sys.target[Y_AXIS] = sys.current_position[Y_AXIS] + FOCUS_OFFSET_Y;
		//
		// Now we can move to machine coord. position 0
		//
		planBufferLine(sys.target[X_AXIS], sys.target[Y_AXIS],HOMING_FEEDRATE/4,OFF,0,0,0);
		st_synchronize();
		planSetPosition(sys.current_position[X_AXIS], sys.current_position[Y_AXIS]);

//		manageInactivity();
		disable_motors();
		// Focus the laser
		if ((sys.current_focus_pos = auto_focus()) == FOCUS_ERROR){
			sendError();
			SERIAL_ERRORLNPGM(MSG_FOCUS_ERROR)
		} else {
			clearToSend();
			SERIAL_ECHOPGM(MSG_FOCUSED);
			SERIAL_ECHOLN((int)sys.current_focus_pos);
			//
			// Move to homing offset position
			//
			sys.target[X_AXIS] = sys.current_position[X_AXIS] + HOMING_OFFSET_X;
			sys.target[Y_AXIS] = sys.current_position[Y_AXIS] + HOMING_OFFSET_Y;
			//
			// Now we can move to machine coord. position 0
			//
			planBufferLine(sys.target[X_AXIS], sys.target[Y_AXIS],HOMING_FEEDRATE/4,OFF,0,0,0);
			st_synchronize();
			// Reset position, now this is the zero
			for(int8_t i=0; i < NUM_AXIS; i++) {
				sys.current_position[i] = 0;
			}
			planSetPosition(sys.current_position[X_AXIS], sys.current_position[Y_AXIS]);

			sys.backlash[X_AXIS]= X_DEFAULT_BACKLASH_MM;
			sys.backlash[Y_AXIS]= Y_DEFAULT_BACKLASH_MM;
		}
	}
}

/**
** clearToSend() - Clear to send
**/
void clearToSend()
{

	SERIAL_PROTOCOLLNPGM(MSG_OK); 
}
/**
** sendError() - Error
**/
void sendError()
{

	SERIAL_PROTOCOLLNPGM(MSG_ERROR); 
}

/**
* clampToSofwareEndstops() 
**/
void clamToSoftwareEnstops(float target[NUM_AXIS])
{
	if (min_software_endstops) {
		if (target[X_AXIS] < X_MIN_POS) target[X_AXIS] = X_MIN_POS;
		if (target[Y_AXIS] < Y_MIN_POS) target[Y_AXIS] = Y_MIN_POS;
	}

	if (max_software_endstops) {
		if (target[X_AXIS] > X_MAX_POS) target[X_AXIS] = X_MAX_POS;
		if (target[Y_AXIS] > Y_MAX_POS) target[Y_AXIS] = Y_MAX_POS;
	}
}


/**
* prepare_line_move() - Prepare the movement of a line command
**/
void prepare_line_move() {

	clamToSoftwareEnstops(sys.target);

	float delta_mm[2], length;
	int laserEntry, laserPower, laserExit;

	delta_mm[X_AXIS] = sys.target[X_AXIS]-sys.current_position[X_AXIS];
	delta_mm[Y_AXIS] = sys.target[Y_AXIS]-sys.current_position[X_AXIS];
	length = sqrt(square(delta_mm[X_AXIS]) + square(delta_mm[Y_AXIS]));

	if (length <= SMALL_SEGMENT_SIZE){
		laserEntry = SMALL_SEGMENT_ENTRY_POWER;             
		laserPower = SMALL_SEGMENT_POWER;
		laserExit  = SMALL_SEGMENT_EXIT_POWER;
	} else {
		laserEntry = DEFAULT_ENTRY_POWER;             
		laserPower = sys.current_laser_power;
		laserExit  = DEFAULT_EXIT_POWER;;               
	} 

	if (sys.current_laser_state == ON)
		planBufferLine(sys.target[X_AXIS], sys.target[Y_AXIS], sys.current_feed_rate,sys.current_laser_state, laserPower, laserEntry, laserExit);
	else
		planBufferLine(sys.target[X_AXIS], sys.target[Y_AXIS], DEFAULT_TRAVELRATE,sys.current_laser_state, laserPower,laserEntry, laserExit);
	update_position();
}
/**
* prepare_arc_move() - Prepare the movement of an arc command
**/
void prepare_arc_move() {
	bool isClockwise;
	float theta;

	theta = radians(numpar[ARC_THETA]);	// The angle of rotation

	//float cordAngle =numpar[SEG_ALPHA];	// The angle of cord
	sys.target[X_AXIS] = numpar[I_COORD] + (sys.current_position[X_AXIS] - numpar[I_COORD]) * cos(theta) - (sys.current_position[Y_AXIS] - numpar[J_COORD]) * sin(theta);
	sys.target[Y_AXIS] = numpar[J_COORD] + (sys.current_position[X_AXIS] - numpar[I_COORD]) * sin(theta) + (sys.current_position[Y_AXIS] - numpar[J_COORD]) * cos(theta);

	sys.offset[X_AXIS] = numpar[I_COORD]-sys.current_position[X_AXIS];		// I (x of center)
	sys.offset[Y_AXIS] = numpar[J_COORD]-sys.current_position[Y_AXIS];		// J (y of center)

	float radius = hypot(sys.offset[X_AXIS], sys.offset[Y_AXIS]); // Compute radius

	if (numpar[ARC_THETA] < 0)	//Angle of rotation
		isClockwise = true;
	else
		isClockwise = false;
	// Trace the arc
	// If error or no movement don't uptade position.
	//
	if (!mc_arc(sys.current_position, sys.target, sys.offset, sys.current_feed_rate, radius, isClockwise,sys.current_laser_state, sys.current_laser_power))
			return;
	update_position();
}

/**
** update_position() - Updates current position
**/
void update_position(){
	sys.current_position[X_AXIS] = sys.target[X_AXIS];
	sys.current_position[Y_AXIS] = sys.target[Y_AXIS];
}


/**
* manageInactivity() - Manage the inactivity of the printer
**/
void manageInactivity() 
{ 
	checkAxesActivity();
}

// Reset to default settings

void config_reset_default()
{
	float tmp1[]=DEFAULT_AXIS_STEPS_PER_UNIT;
	float tmp2[]=DEFAULT_MAX_FEEDRATE;

	memset(&sys, 0, sizeof(sys));  // Clear all system variables

	sys.state = STATE_IDLE;
	sys.execute = 0;

	for (int8_t i=0;i<NUM_AXIS;i++) 
	{
		sys.axis_steps_per_unit[i]=tmp1[i];  
		sys.max_feedrate[i]=tmp2[i];  
		sys.axis_steps_per_sqr_second[i] = DEFAULT_MAX_ACCELERATION * sys.axis_steps_per_unit[i];
		sys.current_position[i] = 0;
		sys.target[i] = 0;
		sys.offset[i]=0;
		sys.backlash[i]= 0.0f;
		sys.current_dir[i] = 1;
	}
	sys.current_feed_rate = DEFAULT_FEEDRATE;
	sys.current_laser_power = DEFAULT_LASER_POWER;
	sys.current_laser_state = false;
	sys.current_focus_pos = 0;
}

// Delays variable defined milliseconds. Compiler compatibility fix for _delay_ms(),
// which only accepts constants in future compiler releases.
void delay_ms(uint16_t ms) 
{
  while ( ms-- ) { _delay_ms(1); }
}


// Delays variable defined microseconds. Compiler compatibility fix for _delay_us(),
// which only accepts constants in future compiler releases. Written to perform more 
// efficiently with larger delays, as the counter adds parasitic time in each iteration.
void delay_us(uint32_t us) 
{
  while (us) {
    if (us < 10) { 
      _delay_us(1);
      us--;
    } else if (us < 100) {
      _delay_us(10);
      us -= 10;
    } else if (us < 1000) {
      _delay_us(100);
      us -= 100;
    } else {
      _delay_ms(1);
      us -= 1000;
    }
  }
}

/**
** main() - The main loop
**/
int main(void) 
{
	// Init serial USB system & reset defaults
	serial_init();
	config_reset_default();

	// init statte and execute a reset to init
	sys.state = STATE_IDLE;  
	sys.execute |= EXEC_RESET;

	for(;;){

		execute_runtime_cmd();
		// If the printer is held don´t accept more HPGL commands
		// only runtime commands
		if (sys.state == STATE_IDLE || sys.state == STATE_ONLINE){
			if(buflen < (BUFSIZE-1))
				getCommand();

			if(buflen)
			{
				process_hpgl_commands(cmdbuffer[bufindr]);
				buflen = (buflen-1);
				bufindr = (bufindr + 1)%BUFSIZE;
			}
		}

		manageInactivity();
		checkHitEndstops();
	}

	return 1;   /* never reached */
}